################################################################################
# Created       : Martin Foster
# Modified      : 01-Jan-2007
################################################################################
#
# User Menu - Script designed to provide consistent structures for signed on
#             users as well as provide alternative displays for those that
#             are not.
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################

package Ethereal::Menu;						# Package name

use strict;							# Strict handling


################################################################################
# Sub-Routines
################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data Members
	my $class    = shift;					# Self reference
	my $database = shift;					# Database handle
	my $param    = shift;					# Parameter handle


	#####################
	# Program Area

	# Create internal hash
	my $self     = {
		DATA   => $database,
		PARAM  => $param,
		HEADER => 'true',
 	 };

	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}


#####################
# Rewrite
#
# Simple and straight forward sub-routine that will take in any type of
# document and correctly implement certain tools and capabilities.

sub Rewrite
{
	#####################
	# Data members
	my $self     = shift;					# Self referencing link
	my $update   = shift;					# Element to update
	my $pass     = shift;					# Passing of parameters

	my $list;						# List of elements
	my $menu     = {};					# Menu based hash

	my $database = $self->{'DATA'};				# Database handler
	my $param    = $self->{'PARAM'};			# Parameter handler


	#####################
	# Program area	

	# Link with hash
	$database->GetHashMenu($menu);

	# Menu displays
	# Authenticated user section
	if ($database->{'AUTH'} eq 'true')
	{
		# Generation of base parameter line
		$pass = $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'})
			unless (defined($pass));	

		#
		# Generation of template
		$list = $menu->{'TmplMenuBase'};

		# Generate specific sections
		# Contributor
		$list .= $menu->{'TmplMenuContrib'}
		  if ($database->{'USR'}{'Verified'} eq 'yes');

		# Supervisory and Moderator
		$list .= $menu->{'TmplMenuSuper'}
		  if (($database->{'USR'}{'SuperPrivs'} eq 'yes')
		   or ($database->{'USR'}{'ModPrivs'} eq 'yes'));

		# Administrative
		$list .= $menu->{'TmplMenuAdmin'}
		  if ($database->{'USR'}{'AdminPrivs'} eq 'yes');


		#
		# Search and replace parameters
		$list =~ s/GENPARAM/$pass/gs;	
	}

	# User is not logged on
	else
	{
		# Simply assign and forget about it
		$list = $menu->{'TmplMenuAuth'};
	}


	# Header management
	# Handle the basics
	if ((defined($database->{'TITLE'}))
	 && ($self->{'HEADER'} eq 'true'))
	{
		# Simple and straight forward update
		$update =~ s/(<TITLE.*?>).+(<\/TITLE.*?>)/${1}$database->{'SYS'}{'TitPrefix'}$database->{'TITLE'}${2}/gsi
	}


	#
	# Search and replace
	$update =~ s/LSTMENU/$list/gs;
	$update =~ s/MSCRIPTTITLE/$database->{'TITLE'}/gs
	  if (defined($database->{'TITLE'}));
	

	# Return updated template
	return $update;
}

1;								# Return true
