#!/usr/bin/perl
use lib '/home/erealms/ethereal/mgmt/perl';

#################################################################################
# Created       : Martin Foster
# Modified      : 19-Mar-2007
#################################################################################
#
# Configure Puppet - Script part of Ethereal Realms, designed to configure 
#                    manage and remove puppets
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################

use CGI qw(-no_debug -nosticky);				# Common gateway interface
use CGI::Carp qw(fatalsToBrowser);				# CGI Error Logs
use Digest::MD5 qw(md5_hex);					# MD5 digest handler					
use Storable qw(nfreeze thaw);					# Storable activity
use strict;							# Strict variable enforcement

use Ethereal::Comm;						# Communications handler
use Ethereal::Database;						# Database handler
use Ethereal::Filter;						# Filter system
use Ethereal::Home;						# Homepage/file handler
use Ethereal::Image;						# Image handler
use Ethereal::Login;						# Login functionality
use Ethereal::Menu;						# Consistent menu
use Ethereal::Option;						# Options handling
use Ethereal::Param;						# Parameter control

#################################################################################
# Gobal override
#################################################################################
$CGI::POST_MAX=1024 * 1024;					# Maximum posts
$CGI::DISABLE_UPLOADS = 1;					# Disable uploads

#################################################################################
# Data Members
#################################################################################
my $cgi;							# Common gateway interface handle
my $database;							# Database handle
my $login;							# Login handle
my $menu;							# Menu handle
my $param;							# Parameter handle
my $tmpl;							# Template handle

my @are;							# Are so cool
my @not;							# Not so cool
my @act;							# Activated puppets
my @sus;							# Suspended puppets
my @pup;							# Puppet handle
my @targets;							# List of targets

my %args;
my %config;							# Configuration tag
my %link;							# Links to directories

#################################################################################
# Program Area
#################################################################################

	# Initial handles
	$cgi      = new CGI;
	$database = new Ethereal::Database();

	# Connect and fetch
	$database->Connect($cgi);

	# Set title
	$database->{'TITLE'} = $database->{'SYS'}{'TitPuppet'};


	# Pre-fetch active and suspended
	@are     = $cgi->param('ARE');
	@not     = $cgi->param('NOT');
	@act     = $cgi->param('APUPPET');
	@sus     = $cgi->param('IPUPPET');
	@pup     = $cgi->param('PUPPET');
	@targets = $cgi->param('TARGET');


	# Parameter handling
	$param = new Ethereal::Param($database, $cgi);
	$param->GetParam();


	# Fetch hash
	$database->GetHashUser(\%config);

	# Create instances
	$tmpl  = new Ethereal::Template(\%config);
	$menu  = new Ethereal::Menu($database, $param);
	$login = new Ethereal::Login($database, $cgi, $param, $menu);

	# Authenthication
	if ($login->GetVerificationNormal())
	{
		# Generate Inline
		# Define base arguments
		%args = (
			USER  => $param->{'USER'},
			CRYPT => $param->{'CRYPT'}
		 );

		# Starting point of necessary
		$args{'PSTART'} = $param->{'PSTART'}
		  if (defined($param->{'PSTART'}));

		# Export inline
		$link{'EXPORT'} = (defined($param->{'ROOM'}))
		  ? $param->EmbedInline(%args, EXPORT=>'True', ROOM=>$param->{'ROOM'})
		  : $param->EmbedInline(%args, EXPORT=>'True');

		# Import inline
		$link{'IMPORT'} = (defined($param->{'ROOM'}))
		  ? $param->EmbedInline(%args, IMPORT=>'True', ROOM=>$param->{'ROOM'})
		  : $param->EmbedInline(%args, IMPORT=>'True');

		# Clone inline
		$link{'CLONE'} = (defined($param->{'ROOM'}))
		  ? $param->EmbedInline(%args, CLONE=>'True', ROOM=>$param->{'ROOM'})
		  : $param->EmbedInline(%args, CLONE=>'True');

		# Group inline
		$link{'GROUP'} = (defined($param->{'ROOM'}))
		  ? $param->EmbedInline(%args, GROUP=>'True', ROOM=>$param->{'ROOM'})
		  : $param->EmbedInline(%args, GROUP=>'True');

		# Activity inline	
		$link{'ACTIVITY'} = (defined($param->{'ROOM'}))
		  ? $param->EmbedInline(%args, ACTIVITY=>'True', ROOM=>$param->{'ROOM'})
		  : $param->EmbedInline(%args, ACTIVITY=>'True');

		# Standard back
		$link{'BACK'} = $param->EmbedInline(%args);
		  
		# Pull URL
		$link{'URL'} = $cgi->url(-absolute=>1);


		# Document header
		print $menu->Rewrite($database->DocumentGetHeader()), "\n";


		# Modal behaviour
		# Normal operations
		unless (defined($database->{'DEFAULT'}))
		{
			# Puppet activity handling
			if (defined($param->{'ACTIVITY'}))
			{ 
				# Take care of business
				if (defined($param->{'MANAGE'}))
				{
					# Will activate suspend as necessary
					ActivityHandler($database, $cgi, $param, \%config, \@act, \@sus); 
				}
	
				# Display information
				ActivityDisplay($database, $cgi, $param, $tmpl, \%config, \%link);
			}

			# Here comes Dolly
			elsif (defined($param->{'CLONE'}))
			{ PuppetClone($database, $cgi, $param, $tmpl, \%config, \%link, \@targets); } 

			# Groupings and tasks
			elsif (defined($param->{'GROUP'}))
			{ PuppetGroup($database, $cgi, $param, $tmpl, \%config, \%link, \@are, \@not); } 

			# Exportable handles
			elsif (defined($param->{'EXPORT'}))
			{ PuppetExport($database, $cgi, $param, $tmpl, \%config, \%link); } 

			# Please have all papers ready and in order
			elsif (defined($param->{'IMPORT'}))
			{ PuppetImport($database, $cgi, $param, $tmpl, \%config, \%link); } 

			# Default behaviour
			else
			{
				# For additions and removal
				unless  ((defined($param->{$config{'TxtView'}})) 
				 || (defined($param->{$config{'TxtSubmit'}})))
				{ InitialDisplay($database, $cgi, $param, $tmpl, \%config, \%link, \@pup); }

				# For puppet changes and viewing
				else
				{ PuppetControl($database, $cgi, $param, $tmpl, \%config, \%link); }
			}
		}

		# Lock-down mode
		else
		{
			# Lock down restricted switch
			$param->{'PUPPET'} = $database->{'DEFAULT'};

			# The one and only choice
			PuppetControl($database, $cgi, $param, $tmpl, \%config, \%link);
		}

		# Document footer	
		print $database->DocumentGetFooter(), "\n";
	}


#################################################################################
# Sub-Routines
#################################################################################

#####################
# Activity Display
#
# Handles the general output of the initial display screen.  Which will allow one
# to change password, E-Mail addresses and/or passwords.

sub ActivityDisplay
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $tmpl     = shift;					# Template handler
	my $config   = shift;					# Configuration hash
	my $link     = shift;					# Hash of links


	my $active;						# Active handles
	my $inactive;						# Inactive handles

	my @active;						# Active handles
	my @inactive;						# Inactive handles
	
	my %args;						# Arguments
	

	#####################
	# Program area	
	

	# Puppet Ignore/Suspend
	# Populate arrays
	$database->GetListPuppetYours(\@active, $param->{'USER'});
	$database->GetListPuppetSuspended(\@inactive, $param->{'USER'});

	# Format lists
	$active   = $cgi->scrolling_list('APUPPET', \@active,   $active[0],   5, 'true');
	$inactive = $cgi->scrolling_list('IPUPPET', \@inactive, $inactive[0], 5, 'true');

	# Form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), MANAGE=>'True');

	# Search and replace template
	$tmpl->Show('TmplPupActivity',
                LACTVT   => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
		LBACK    => "HREF=\"$link->{URL}$link->{BACK}\"",
                LCLONE   => "HREF=\"$link->{URL}$link->{CLONE}\"",
                LEXPRT   => "HREF=\"$link->{URL}$link->{EXPORT}\"",
                LIMPRT   => "HREF=\"$link->{URL}$link->{IMPORT}\"",
		LGROUP   => "HREF=\"$link->{URL}$link->{GROUP}\"",
		WACTIVE  => $active,
		WDISABLE => $inactive,
		WPUPOFF  => $cgi->submit($config->{'TxtSuspend'}),
		WPUPON   => $cgi->submit($config->{'TxtActive'})
	 );

	
	# Embed
	print $cgi->end_form(), "\n";
}


#####################
# Activity Handler
#
# Will handle the activation de-activation of puppets as designated by the user
# this is to reduce the long list of puppets for selection when some are rarely
# used if ever.

sub ActivityHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $config   = shift;					# Configuration tags	

	my $act      = shift;					# Active list
	my $sus      = shift;					# Suspended list

	my $active;						# Activity flag
	my $puppet;						# Puppet to handle

	my @puppets;						# Puppets to handle


	#####################
	# Program area	

	# Find appropraite values
	$active  = (defined($param->{$config->{'TxtActive'}})) ? 'NULL'  : 'LOCALTIMESTAMP'; 
	@puppets = (defined($param->{$config->{'TxtActive'}})) ? @{$sus} : @{$act};

	# Update entries
	eval
	{
		# Loop
		foreach $puppet (@puppets)
		{
			# Update activity
			$database->Write("UPDATE Puppet
				SET PuppetActive=$active
				WHERE PuppetName=?
				  AND PuppeteerLogin=?",

			 $puppet,
			 $param->{'USER'}
			);
		}

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Failed to toggle activity - Action aborted: $@");

		# Undo incomplete changes
		$database->Rollback();
	}

	# Cleanup
	$param->Cleanup('PUPPET');
}


#####################
# Conflict management
#
# Occasionally an imported handle will conflic with existing sets
# this will essentially find a variant name that will work
# properly.

sub Conflict
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $param    = shift;					# CGI handle
	my $char     = shift;					# Passed character

	my $exists;						# Existence
	my $handle;						# Handle
	my $rand;						# Random number

	my $short;						# Short name
	my $test;						# Tested short name

	#####################
	# Program area

	# Initialize
	$handle = $char;

	# Generate short name
	$short  =  lc($handle);
	$short  =~ s/-/_/gs;
	$short  =~ s/\s/_/gs;
	$short  =~ s/\W//gs;
	
	# Only if uniqueness is enforced
	if (defined($param->{'IMPORT_UNIQUE'}))
	{
		# Initial check
		($exists) = $database->DataGetPuppetCreate($handle, $short, $param->{'USER'});

		# Continued checks
		while (defined($exists))
		{
			# Generate number
			$rand = int(rand() * 1000);
			
			# Generate variant
			$handle = $char  . ' ' . $rand;
			$test   = $short . '_' . $rand;
		
			# Does it exist
			($exists) = $database->DataGetPuppetCreate($handle, $test, $param->{'USER'});
		}

		# Transfer over if necessary
		$short = $test if (defined($test));
	}

	# Simply run a check
	else
	{
		# Initial check
		($exists) = $database->DataGetPuppetCreate($handle, $short, $param->{'USER'});

		# Return undefined if exists
		return undef if (defined($exists));
	}

	# Return
	return $handle, $short;
}


#####################
# Image Legal
#
# Determines if an image is actually legal based on a regular expression search
# through a list of known banned images.

sub ImageLegal
{
	####################
	# Data Members

	my $database = shift;					# Database handle	
	my $link     = shift;					# Link to image

	my $ban;						# Specific ban
	my @bans;						# List of bans


	####################
	# Program Area

	# Get list of banned images
	$database->GetListImageBan(\@bans);


	# Cycle through
	foreach $ban (@bans)
	{
		# Verify against link
		if ($link =~ /^$ban/si)
		{
			# Return illegal
			return 0;
		}		
	}


	# Return legal
	return 1;
}

#####################
# Image Size
#
# Downloads a picture off the net and determins it's size as to be able place 
# the defined sizes on the database

sub ImageSize
{
        ####################
        # Data Members

	my $database = shift;					# Database handle	
        my $link     = shift;					# URL to picture

        my $file;						# Name of file
	my $home;						# Homepage handler
	my $image;						# Image handle

	my $width;						# Image width
	my $height;						# Image height
	my $size;						# Image size


        ####################
        # Program Area

	# Initiate Image and Homepage/File handles
	$image = new Ethereal::Image($database, $database->{'SYS'}{'LocTemp'});
	$home  = new Ethereal::Home($database, $database->{'SYS'}{'LocTemp'});


	# Fetch image
	($file) = $home->FileFetch($link);

	
	# Retreive demensions
	($width,
	 $height,
         $size) = $image->Size($file);

	# Unlink
	$home->FileUnlink($file); 


	# Return values
	return $width, $height, $size;
}


#####################
# InitialDisplay
#
# Handles the general output of the initial display screen.  Which will allow one
# to change password, E-Mail addresses and/or passwords.

sub InitialDisplay
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $link     = shift;					# Hash of links
	my $pup      = shift;					# Puppet list

	my $audit;						# Auditing handle
	my $filter;						# Filter handle

	my $list;						# List of puppets
	my $conf;						# Confirmation
	my $newchar;						# New character
	my $clone;						# Cloning

	my $handle;						# Handle name
	my $short;						# Associated short name

	my $buttonmake;						# Create button
	my $buttonrem;						# Button remove
	my $buttonview;						# Button view

	my $tag;						# Tagline
	my $title;						# Title
	my $count;						# Active puppets

	my $choice;						# Choice of the matter
	my $message;						# Display message

	my @puppets;						# Array of puppets


	#####################
	# Program area	
	
	# Create instance
	$audit    = new Ethereal::Audit($database, $cgi, $param);
	$filter   = new Ethereal::Filter();

	# Evaluation
	eval
	{
		# Addition operation
		if ( defined($param->{$config->{'TxtAddition'}}) ) 
		{
			# Strip HTML
			$handle =  $filter->StripHTML($param->{'PUPADD'});

			# Additional filtering
			$handle =~ s/~//gs;

			# Generate short name
			$short  =  lc($handle);
			$short  =~ s/-/_/gs;
			$short  =~ s/\s/_/gs;
			$short  =~ s/\W//gs;

			# Retreive information
			($choice) = $database->DataGetPuppetCreate($handle, $short, $param->{'USER'});


			# Verification
			if ((length($param->{'PUPADD'}) > 2)
			 && ($param->{'PUPADD'} !~ /^\s/)
			 && (!defined($choice)))
			{
				# Tagline handling
				# Pull taglines
				if (defined($param->{'PUPCLONE'}))
				{
					# Pull from database
					($title, $tag) = $database->DataGetPuppetFormat(
						$param->{'PUPPET'}, 
						$param->{'USER'}
					 );
				}

				# Insertion upon acceptance
				$database->Write("INSERT INTO Puppet 
					(PuppetName, 
       	                          	 PuppeteerLogin,
					 PuppetShort,
					 PuppetTitle,
					 PuppetTag)
					VALUES(?,?,?,?,?)",

				 $handle, 
				 $param->{'USER'},
				 $short,
				 $title,
				 $tag);

				# Record change
				$audit->PuppetAdd($param->{'PUPADD'});


				# Display status
				$message = $config->{'MsgCreate'};
			}
			else
			{ $message = $config->{'MsgExists'}; }
		}

		# Removal operation
		if ((defined($param->{$config->{'TxtDelete'}}))
		 && (defined($param->{'CONFIRM'})))
		{
			# Loop through all handles
			foreach my $puppet (@{$pup})
			{
				# Retreive information
				($choice) = $database->DataGet('SELECT PuppetName
					FROM Puppet 
					WHERE PuppetName=? 
					  AND PuppeteerLogin=?', 

				 $puppet, 
				 $param->{'USER'}
				);


				# Verification
				if (defined($choice))
				{
					# Removal of primary tables
					$database->Write("DELETE FROM Puppet 
						WHERE PuppetName=?
						  AND PuppeteerLogin=?", 

					 $puppet,
					 $param->{'USER'}
					);

					# Record change
					$audit->PuppetRem($puppet);


					# Message of success
					$message = $config->{'MsgDeletePass'}; 
				}		
				else
				{ $message = $config->{'MsgDeleteFail'}; }
			}
		}

		# Commit
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Puppet creation/removal aborted: $@");

		# Undo incomplete changes
		$database->Rollback();
	}


	# Cleanup
	$param->Cleanup('CONFIRM','PUPCLONE','PUPHTML');


	# Display of values
	# Retreive a list of puppets
	$database->GetListPuppetComplete(\@puppets, $param->{'USER'});

	# Total
	$count = @puppets;


	# Buttons
	# Default arguments



	# Define message
	$message = (defined($message)) ? $message : '';


	# Prepare widgets
	$list    = $cgi->scrolling_list('PUPPET', \@puppets, $puppets[0], 10, 'true');
	$newchar = $cgi->textfield('PUPADD', '', 25, 30);

	# Checkboxes
	$conf    = $cgi->checkbox(-name=>'CONFIRM', -label=>'');
	$clone   = $cgi->checkbox(-name=>'PUPCLONE', -label=>'');

	# Buttons
	$buttonview = $cgi->submit($config->{'TxtView'});
	$buttonrem  = $cgi->submit($config->{'TxtDelete'});
	$buttonmake = $cgi->submit($config->{'TxtAddition'});


	# Selection/removal
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";


	# Search and replace template
	$tmpl->Show('TmplPupList',
                LACTVT => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
		LBACK  => "HREF=\"$link->{URL}$link->{BACK}\"",
                LCLONE => "HREF=\"$link->{URL}$link->{CLONE}\"",
                LEXPRT => "HREF=\"$link->{URL}$link->{EXPORT}\"",
                LIMPRT => "HREF=\"$link->{URL}$link->{IMPORT}\"",
		LGROUP => "HREF=\"$link->{URL}$link->{GROUP}\"",
		MSTATS => $message,
		MCOUNT => $count,
		WCLNE  => $clone,
		WCONF  => $conf,
		WLIST  => $list,
		WNEWC  => $newchar,
		WDELE  => $buttonrem,
		WMAKE  => $buttonmake,
		WVIEW  => $buttonview
	 );


	# End form
	print $cgi->end_form();
}


#####################
# PuppetClone
#
# Allows the cloning (as in the sheep) of tagline and titles from one handle
# to others.  Basically a straight forward process and may speed up mass
# changes. 

sub PuppetClone
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $link     = shift;					# Hash of links
	my $targets  = shift;					# Array of targets

	my $option   = new Ethereal::Option();			# Option handle

	my $mode;						# Copy mode
	my $puppet;						# Puppets selection
	my $target;						# List of targets

	my $inline;						# Inline parameter handling
	my $url;						# Return link

	my $count;						# Counter
	my $tag;						# Tagline
	my $title;						# Title
	my $message;						# Display message
	my $placeholders;					# Placeholders

	my @modes;						# Clone modes
	my @puppets;						# Array of puppets


	#####################
	# Program area	

	# Clone handling
	# Only if necessary
	if ((defined($param->{$config->{'TxtClone'}}))
	 && (defined($param->{'PUPPET'})))
	{
		# Placeholders
		# Fetch total
		$count = @{$targets};	

		# Generate placeholders
		$placeholders = "?," x $count;
		chop($placeholders);


		# Pull tagline and tite
		($title, $tag) = $database->DataGetPuppetFormat(
			$param->{'PUPPET'}, 
			$param->{'USER'}
		 );

		# Discover mode
		$mode = $option->FindClone($config->{'OptClone'}, $param->{'MODE'});


		# Operate based on mode
		if ($count > 0)
		{
			# Tagline only
			if ($mode eq 'tag')
			{
				# Update of tagline only
				$database->Quick("UPDATE Puppet
					SET PuppetTag=?
					WHERE PuppeteerLogin=?
					  AND PuppetName IN ($placeholders)",

				 $tag,
				 $param->{'USER'},
				 @{$targets}
				);
			}

			# Title
			elsif ($mode eq 'title')
			{
				# Update of title only
				$database->Quick("UPDATE Puppet
					SET PuppetTitle=?
					WHERE PuppeteerLogin=?
					  AND PuppetName IN ($placeholders)", 

				 $title,
				 $param->{'USER'},
				 @{$targets}
				);
			}

			# Both title and tag
			elsif ($mode eq 'both')
			{
				# Update of tagline only
				$database->Quick("UPDATE Puppet
					SET PuppetTag=?,
					    PuppetTitle=?
					WHERE PuppeteerLogin=?
					  AND PuppetName IN ($placeholders)",

				 $tag,
				 $title,
				 $param->{'USER'},
				 @{$targets}
				);
			}

			# Warm and fuzzy message
			$message = $config->{'MsgClone'};
		}

		else
		{
			# Tap on wrist
			$message = $config->{'MsgCount'};
		}

		# Cleanup
		$param->Cleanup($config->{'TxtClone'});
	}

	
	# Pull valies
	# Retreive a list of puppets
	$database->GetList(\@puppets, "SELECT PuppetName 
		FROM Puppet 
		WHERE PuppeteerLogin=? 
		  AND PuppetActive IS NULL
		ORDER BY PuppetName", 

	 $param->{'USER'}
	);

	# Mode
	$option->Split(\@modes, $config->{'OptClone'});

	# Back button
	# Generate Inline and URL
	$inline = (defined($param->{'ROOM'}))
	  ? $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'}, ROOM=>$param->{'ROOM'})
	  : $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'});	
	$url    = $cgi->url(-absolute=>1);


	# Prepare widgets
	$mode   = $cgi->radio_group('MODE', \@modes, $modes[2], 'true');
	$puppet = $cgi->scrolling_list('PUPPET', \@puppets, $puppets[0], 5);
	$target = $cgi->scrolling_list('TARGET', \@puppets, $puppets[0], 5, 'true');


	# Information display
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Search and replace template
	$tmpl->Show('TmplPupClone',
                LACTVT   => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
		LBACK    => "HREF=\"$link->{URL}$link->{BACK}\"",
                LCLONE   => "HREF=\"$link->{URL}$link->{CLONE}\"",
                LEXPRT   => "HREF=\"$link->{URL}$link->{EXPORT}\"",
                LIMPRT   => "HREF=\"$link->{URL}$link->{IMPORT}\"",
		LGROUP   => "HREF=\"$link->{URL}$link->{GROUP}\"",
		MMESSAGE => $message,
		WPUPPET  => $puppet,
		WTARGET  => $target,
		WMODE    => $mode,
		WCLONE   => $cgi->submit($config->{'TxtClone'})
	 );

	# End form
	print $cgi->end_form();
}


#####################
# PuppetGroup
#
# Will permit users to group their puppets into different categories based
# on their needs.   This would only show up in realms themselves, as other
# places do not need that level of complexity.

sub PuppetGroup
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $link     = shift;					# Hash of links
	my $are      = shift;					# Generated in group
	my $not      = shift;					# Generated lesser then group


	my $control;						# Control widget
	my $count;						# Number of groups
	my $group;						# Singular group
	my $puppet;						# Individual puppet

	my $groups;						# Grouping links
	my $ware;						# Generated in group
	my $wnot;						# Generated lesser then group

	my @are;						# Are part of the group
	my @not;						# Not part of the group
	my @groups;						# List of groups


	#####################
	# Program area	

	# Information handling
	# Creation
	if ((defined($param->{$config->{'TxtAddition'}}))
	 && (defined($param->{'NEWGROUP'}))
	 && ($param->{'NEWGROUP'} =~ /^\w{3,10}$/))
	{
		# Create new group
		$database->Quick("INSERT INTO Association
			(PuppeteerLogin,
			 AssociationName)
			VALUES(?,?)",

		 $param->{'USER'},
		 $param->{'NEWGROUP'}
		);

		# Change current group
		$param->{'CURRENT'} = $param->{'NEWGROUP'};

		# Cleanup
		$param->Cleanup('NEWGROUP');
	}

	# Removal
	elsif ((defined($param->{$config->{'TxtDelete'}}))
	  && (defined($param->{'CURRENT'})))
	{
		# Evaluate operation
		eval
		{
			# Remove from association
			$database->Write("DELETE FROM Association
				WHERE PuppeteerLogin=?
				  AND AssociationName=?",

			 $param->{'USER'},
			 $param->{'CURRENT'}
			);

			# Make sure no puppets carry this
			$database->Write("UPDATE Puppet
				SET PuppetAssociation=NULL
				WHERE PuppeteerLogin=?
				  AND PuppetAssociation=?",

			 $param->{'USER'},
			 $param->{'CURRENT'}
			);

			# Commit
			$database->Commit();

			# Cleanup
			$param->Cleanup('CURRENT');
		};

		# Operation failure
		if ($@)
		{
			# Issue warning
			warn("Group removal aborted: $@");

			# Undo incomplete changes
			$database->Rollback();
		}
	}

	# Puppet updates
	# Additions
	elsif ((defined($param->{$config->{'TxtMoveIn'}}))
	  && (defined($param->{'CURRENT'})))
	{
		# Evaluate
		eval
		{
			# Loop through
			foreach $puppet (@{$not})
			{
				# Update puppets group
				$database->Write("UPDATE Puppet
					SET PuppetAssociation=?
					WHERE PuppeteerLogin=?
					  AND PuppetName=?",

				 $param->{'CURRENT'},
				 $param->{'USER'},
				 $puppet
				);
			}

			# Commit changes
			$database->Commit();
		};

		# Operation failure
		if ($@)
		{
			# Issue warning
			warn("Group member addition aborted: $@");

			# Undo incomplete changes
			$database->Rollback();
		}
	}
                
	# Removal
	elsif ((defined($param->{$config->{'TxtMoveOut'}}))
	  && (defined($param->{'CURRENT'})))
	{
		# Evaluate
		eval
		{
			# Loop through
			foreach $puppet (@{$are})
			{
				# Update puppets group
				$database->Write("UPDATE Puppet
					SET PuppetAssociation=NULL
					WHERE PuppeteerLogin=?
					  AND PuppetName=?",

				 $param->{'USER'},
				 $puppet
				);
			}

			# Commit changes
			$database->Commit();
		};

		# Operation failure
		if ($@)
		{
			# Issue warning
			warn("Group member exclusion aborted: $@");

			# Undo incomplete changes
			$database->Rollback();
		}
	}
                

	# Pull values
	# Retreive a list of puppets
	$database->GetListGroup(\@groups, $param->{'USER'});

	# Number of groups
	$count = @groups;

	# Information display
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";


	# This user is picky or has too many handles
	if ($count >= 1)
	{
		# Ensure definition
		$param->{'CURRENT'} = (defined($param->{'CURRENT'})) ? $param->{'CURRENT'} : $groups[1];

		# Pull lists
		# Not associated
		$database->GetList(\@not, "SELECT PuppetName
			FROM Puppet
			WHERE PuppeteerLogin=?
			  AND (PuppetAssociation <> ?
			   OR PuppetAssociation IS NULL)
			ORDER By PuppetName",

		 $param->{'USER'},
		 $param->{'CURRENT'}
		);
		
		# Associated
		$database->GetList(\@are, "SELECT PuppetName
			FROM Puppet
			WHERE PuppeteerLogin=?
			  AND PuppetAssociation=?
			ORDER BY PuppetName",

		 $param->{'USER'},
		 $param->{'CURRENT'}
		);
		
		# Generated widgets
		$ware = $cgi->scrolling_list('ARE', \@are, $are[0], 5, 'true');
		$wnot = $cgi->scrolling_list('NOT', \@not, $not[0], 5, 'true');

		# Control widget
		$control = $cgi->submit($config->{'TxtMoveIn'}) . "<BR><BR>" . $cgi->submit($config->{'TxtMoveOut'});

		# Initialize
		$groups = '';
		
		# Create group links
		foreach $group (@groups)
		{
			# Display new link
			$groups .= $tmpl->Pass('TagTabbed', 
				MLINK=>"$link->{'URL'}$link->{'GROUP'}&CURRENT=$group",
				MNAME=>$group);
		}

		# Search and replace template
		$tmpl->Show('TmplPupGroups',
			LACTVT   => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
			LBACK    => "HREF=\"$link->{URL}$link->{BACK}\"",
			LCLONE   => "HREF=\"$link->{URL}$link->{CLONE}\"",
			LEXPRT   => "HREF=\"$link->{URL}$link->{EXPORT}\"",
			LIMPRT   => "HREF=\"$link->{URL}$link->{IMPORT}\"",
			LGROUP   => "HREF=\"$link->{URL}$link->{GROUP}\"",
			MCURRENT => ucfirst($param->{'CURRENT'}),
			MGROUPS  => $groups,
			WNOT     => $wnot,
			WCTRL    => $control,
			WARE     => $ware,
			WDELETE  => $cgi->submit($config->{'TxtDelete'}),
			WCREATE  => $cgi->submit($config->{'TxtAddition'}),
			WGROUP   => $cgi->textfield('NEWGROUP', '', 10, 10)
		 );
	}

	# Generic template
	else
	{
		# Search and replace template
		$tmpl->Show('TmplPupGroupsN',
			LACTVT   => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
			LBACK    => "HREF=\"$link->{URL}$link->{BACK}\"",
			LCLONE   => "HREF=\"$link->{URL}$link->{CLONE}\"",
			LEXPRT   => "HREF=\"$link->{URL}$link->{EXPORT}\"",
			LIMPRT   => "HREF=\"$link->{URL}$link->{IMPORT}\"",
			LGROUP   => "HREF=\"$link->{URL}$link->{GROUP}\"",
			WCREATE  => $cgi->submit($config->{'TxtAddition'}),
			WGROUP   => $cgi->textfield('NEWGROUP', '', 10, 10)
		 );
	}

	
	# End form
	print $cgi->end_form();
}


#####################
# Puppet Export
#
# Creates a human readable file with all that is necessary to actually 
# move a handle somewhere else.   Or create backups for the paranoid.

sub PuppetExport
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $link     = shift;					# Hash of links

	my $textarea;						# Textarea widget
	my $export;						# Exportable hash
	my $wrapped;						# Wrapped text

	my $count      = 0;					# Counter of position
	my $wrap       = 72;					# Wrap to how much
	
	my $res;						# Results handle
	my $statement;						# Query handle
	
	my %char;						# Hash holding all characters
	
	#####################
	# Program hash
	
	# Prepare query
	$database->Pull(\$statement, "SELECT
		  PuppetName        AS \"PuppetName\",
		  PuppetDefault     AS \"PuppetDefault\",
		  PuppetBio         AS \"PuppetBio\",
		  PuppetTag         AS \"PuppetTag\",
		  PuppetTitle       AS \"PuppetTitle\",
		  PuppetPic         AS \"PuppetPic\",
		  PuppetPicHeight   AS \"PuppetPicHeight\",
		  PuppetPicWidth    AS \"PuppetPicWidth\",
		  PuppetPicSize     AS \"PuppetPicSize\",
		  PuppetTextColour  AS \"PuppetTextColour\",
		  PuppetTextFace    AS \"PuppetTextFace\",
		  PuppetAssociation AS \"PuppetAssociation\"
		FROM Puppet
		WHERE PuppeteerLogin=?
		ORDER BY PuppetName",

	 $param->{'USER'}
 	);

	# Run through and populate hash
	while ($res = $statement->fetchrow_hashref())
	{
		# Add entry to hash
		# Additional entries themselves hashes
		$char{$res->{'PuppetName'}} = {
		  PuppetDefault     => $res->{'PuppetDefault'},
		  PuppetBio         => $res->{'PuppetBio'},
		  PuppetTag         => $res->{'PuppetTag'},
		  PuppetTitle       => $res->{'PuppetTitle'},
		  PuppetPic         => $res->{'PuppetPic'},
		  PuppetPicHeight   => $res->{'PuppetPicHeight'},
		  PuppetPicWidth    => $res->{'PuppetPicWidth'},
		  PuppetPicSize     => $res->{'PuppetPicSize'},
		  PuppetTextColour  => $res->{'PuppetTextColour'},
		  PuppetTextFace    => $res->{'PuppetTextFace'},
		};

		# No need if undefined
		$char{$res->{'PuppetName'}}->{'PuppetAssociation'} = $res->{'PuppetAssociation'}
	          if (defined($res->{'PuppetAssociation'}));
	}

	# Complete statement
	$statement->finish();


	# Create exportable string
	# Package the hash and encode
	$export = unpack('H*', nfreeze(\%char));

	# Rewrap
	# Initialize
	$wrapped = '';

	# Cut through
	for ($count=0; $count <= (int(length($export) / $wrap)); $count++)
	{
		# Cut portion of text
		$wrapped .= substr($export, ($count*$wrap), $wrap) . "\n";
	}
	

	# Display prep
	# Generate textarea
	$textarea = $cgi->textarea(
		-name    => 'EXPORT_PUPPET',
		-rows    => 4,
		-columns => 75,
		-wrap    => 'SOFT',
		-default => $wrapped
	);

	# Form
	print $cgi->start_form();
	
	# Display template
	$tmpl->Show('TmplPupExport',
		LACTVT     => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
		LBACK      => "HREF=\"$link->{URL}$link->{BACK}\"",
		LCLONE     => "HREF=\"$link->{URL}$link->{CLONE}\"",
		LEXPRT     => "HREF=\"$link->{URL}$link->{EXPORT}\"",
		LIMPRT     => "HREF=\"$link->{URL}$link->{IMPORT}\"",
		LGROUP     => "HREF=\"$link->{URL}$link->{GROUP}\"",
		WPUPEXPORT => $textarea,
	 );

	# End form
	print $cgi->end_form(), "\n";
}


#####################
# Puppet Import
#
# Makes use of exported text, rids itself of any non-natural characters
# and then converts into a datatype useable for import.

sub PuppetImport
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $link     = shift;					# Hash of links

	my $audit;						# Audit handle

	my $assoc    = 'PuppetAssociation';			# Row with significant meaning
	my $exists;						# Existence of association
	
	my $choice;						# Handle exists
	my $action;						# Coded SQL action
	my $place;						# Placeholders

	my $count;						# Internal counter
	my $char;						# Character
	my $import;						# Import line
	my $key;						# Coded key
	my $list;						# List of successes
	my $name;						# Official name
	my $short;						# Short name
	my $textarea;						# Textarea widget
	
	my @attr;						# Attributes
	
	my %char;						# Character line
	my %groups;						# Hash of associations
	my %sanctioned = (					# Sanctioned hashes
		PuppetName        => 1,
		PuppetDefault     => 1,
		PuppetBio         => 1,
		PuppetTag         => 1,
		PuppetTitle       => 1,
		PuppetPic         => 1,
		PuppetPicHeight   => 1,
		PuppetPicWidth    => 1,
		PuppetPicSize     => 1,
		PuppetTextColour  => 1,
		PuppetTextFace    => 1,
		PuppetAssociation => 1 
	 );
	

	#####################
	# Program hash
	
	# Initialize list
	$list = '';
	
	# Database handling
	if (defined($param->{'IMPORT_PUPPET'}))
	{
		# Massage
		$import =  $param->{'IMPORT_PUPPET'};
		$import =~ s/(\s|\n|$)//gs;

		# Attempt to recover
		eval
		{
			# Recover line
			%char = %{thaw(pack('H*', $import))};
		};

		# No errors continue
		unless ($@)
		{
			# Create instance
			$audit    = new Ethereal::Audit($database, $cgi, $param);
			
			# Cycle through
			foreach $char (keys(%char))
			{
				# Conflict management
				($name, $short) = Conflict($database, $param, $char);
				
				# Skip
				next unless (defined($name));

				# Reset
				undef @attr;	

				# Initialize
				$count  = 0;
				
				# Begin string
				$action = "INSERT INTO Puppet (";


				# Cycle through fields
				foreach $key (keys(%{$char{$char}}))
				{
					# Skip if necessary additional unknown tags
					next unless (defined($sanctioned{$key}));


					# Increment
					$count++;

					# Append field of name
					$action .= $key . "\,";

					# Append onto array
					push(@attr, $char{$char}->{$key});


					# Association addition
					if ($key eq $assoc)
					{
						# Reduce server side queries
						unless (defined($groups{$char{$char}->{$key}}))
						{
							# Query for existence
							($exists) = $database->DataGet("SELECT AssociationName
								FROM Association
								WHERE PuppeteerLogin=?
								  AND AssociationName=?",

							 $param->{'USER'},
							 $char{$char}->{$key}
						 	);


							# Create association if necessary
							$database->Quick("INSERT INTO Association
								(PuppeteerLogin,
			 					 AssociationName)
								VALUES(?,?)",

							 $param->{'USER'},
							 $char{$char}->{$key}

							) unless (defined($exists));


							# Create hash entry
							$groups{$char{$char}->{$key}} = 0;
						}


						# Increment
						$groups{$char{$char}->{$key}}++;
					}
				}

				# Two required fields
				$action .= "PuppetName,PuppetPool,PuppeteerLogin,PuppetShort";

				# Generate placeholders
				$place  = '?,' x $count;
				

				# Complete line
				$action .= ") VALUES(" . $place . "?,'system',?,?)";
				

				# Push additional two
				push(@attr, $name, $param->{'USER'}, $short);
				

				# Individual tries
				eval
				{
					# Write to database
					$database->Write($action, @attr);

					# Record change
					$audit->PuppetAdd($char);

					# Commit change
					$database->Commit();

					# Append to list
					$list .= $tmpl->Pass('TagPupImportList', MCHAR=>$name);					
				};

				# Failure
				if ($@)
				{
					# Roll back transaction
					$database->Rollback();
				}
			}
		}
	}
	
	# Cleanup
	$param->Cleanup('IMPORT_UNIQUE', 'IMPORT_PUPPET');
	

	# Generate textarea
	$textarea = $cgi->textarea(
		-name    => 'IMPORT_PUPPET',
		-rows    => 4,
		-columns => 75,
		-wrap    => 'SOFT',
		-default => ''
	);

	# Form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";
	
	# Display template
	$tmpl->Show('TmplPupImport',
		LACTVT     => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
		LBACK      => "HREF=\"$link->{URL}$link->{BACK}\"",
		LCLONE     => "HREF=\"$link->{URL}$link->{CLONE}\"",
		LEXPRT     => "HREF=\"$link->{URL}$link->{EXPORT}\"",
		LIMPRT     => "HREF=\"$link->{URL}$link->{IMPORT}\"",
		LGROUP     => "HREF=\"$link->{URL}$link->{GROUP}\"",
		LSTIMPORT  => $list,
		WPUPIMPORT => $textarea,
		WPUPSUBMIT => $cgi->submit($config->{'TxtImport'}),
		WPUPUNIQUE => $cgi->checkbox(-name=>'IMPORT_UNIQUE', -label=>'', -checked=>'on'),
		WPUPRESET  => $cgi->reset($config->{' TxtReset'})
	 );

	# End form
	print $cgi->end_form(), "\n";
}


#####################
# PuppetControl
#
# Sub-routine that will handle the change of information for a puppet.  This
# is assuming that the submit button was pressed.

sub PuppetControl
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $link     = shift;					# Hash of links

	my $statement;						# Query handle

	my $active;						# Active choice
	my $default;						# Default handle
	my $preview  = '';					# Room display
	my $title;						# Title bar
	my $tag;						# Tagline bar
	my $wtitle;						# Title bar
	my $wtag;						# Tagline bar

	my $s_active;						# Activity signature
	my $s_default;						# Default handle signature

	my $choice   = 0;					# Choice of the matter
	my $info;						# Puppet information
	my $gummies;						# Gummies html
	my $gummy;						# Default entry
	my $sig;						# Signature block

	my @gummies;						# List of gummies


	#####################
	# Program area

	# Change of information and verify for appropriate button press
	if ( defined($param->{$config->{'TxtSubmit'}}) )
	{ 
		# Basic check
		# Ensure correctness
		$s_active  = (defined($param->{'PUPACT'})) ? 'yes' : 'no';
		$s_default = (defined($param->{'PUPDEF'})) ? 'yes' : 'no';

		# Generate signature
		$sig = md5_hex(
			$param->{'PUPIMG'} .
		 	$param->{'PUPBIO'} .
			$param->{'GUMMY'} .
			$s_active .
			$s_default .
			$param->{'PUPTIT'} .
			$param->{'PUPTAG'}
		 );

		# No match change
		PuppetControlWrite($database, $param, $config)
		  if ($param->{'MD5_SIG'} ne $sig); 
	}


	# Gather information
	# Prepare and execute
	$database->Pull(\$statement, "SELECT
			 PuppetName       AS \"PuppetName\",
			 PuppetActive     AS \"PuppetActive\",
			 PuppetBio        AS \"PuppetBio\",
			 PuppetDefault    AS \"PuppetDefault\",
			 PuppetGummy      AS \"PuppetGummy\",
			 PuppetPic        AS \"PuppetPic\",
			 PuppetPicHeight  AS \"PuppetPicHeight\",
			 PuppetPicSize    AS \"PuppetPicSize\",
			 PuppetPicWidth   AS \"PuppetPicWidth\",
			 PuppetPool       AS \"PuppetPool\",
			 PuppetTag        AS \"PuppetTag\",
			 PuppetTextColour AS \"PuppetTextColour\",
			 PuppetTextSize   AS \"PuppetTextSize\",
			 PuppetTitle      AS \"PuppetTitle\"
			FROM  Puppet
			WHERE PuppetName=? 
			  AND PuppeteerLogin=?",

	 $param->{'PUPPET'},
	 $param->{'USER'}
	);

	# Retreive into hash
	$info = $statement->fetchrow_hashref();


	# Gummy list
	# Retreive icons general or canon
	if ($info->{'PuppetPool'} eq 'system')
	{
		# Default and general
		GetGummyDefault($database, \@gummies);
		GetGummyGeneral($database, \@gummies);
	}
	else
	{
		# Simply canon handles
		GetGummyCanon($database, \@gummies, $param->{'ROOM'});
	}

	# Administrative, Supservisory and Moderator gummies
	GetGummyContrib($database, \@gummies) if ($database->{'USR'}{'Verified'} eq 'yes');
	GetGummyAdmin($database, \@gummies)   if ($database->{'USR'}{'AdminPrivs'} eq 'yes');
	GetGummySuper($database, \@gummies, $param->{'ROOM'}) if ($database->{'USR'}{'SuperPrivs'} eq 'yes');
	GetGummyMod($database, \@gummies, $param->{'ROOM'})   if ($database->{'USR'}{'ModPrivs'} eq 'yes');

	# None option
	push(@gummies, $config->{'TxtNone'})  if ($database->{'USR'}{'Verified'} eq 'yes');


	# Reverse lookup gummy
	($gummy)  = ($info->{'PuppetGummy'} ne 'NIL')
	  ? $database->DataGetGummyName($info->{'PuppetGummy'})
	  : $config->{'TxtNone'};

	# Determin default and format
	$gummies  = $cgi->popup_menu('GUMMY', \@gummies, $gummy);


	# Add hint for users
	$info->{'PuppetPic'} = (defined($info->{'PuppetPic'})) ? $info->{'PuppetPic'} : 'http://';
	$info->{'PuppetBio'} = (defined($info->{'PuppetBio'})) ? $info->{'PuppetBio'} : 'http://';


	# Display of information
	# Javascript
	print "$config->{SetFormat}\n\n"
	  if ($database->{'USR'}{'Javascipt'} eq 'yes');


	# Active handles
	# Handle is active
	if (!defined($info->{'PuppetActive'}))
	{
		$active   = $cgi->checkbox(-name=>'PUPACT', -checked=>'on', -label=>'');
		$s_active = 'yes';
	}

	# Handle is suspended
	else
	{
		$active   = $cgi->checkbox(-name=>'PUPACT', -label=>'');
		$s_active = 'no';
	}


	# Default handles
	# Handle is active
	if ($info->{'PuppetDefault'} eq 'yes')
	{
		$default   = $cgi->checkbox(-name=>'PUPDEF', -checked=>'on', -label=>'');
		$s_default = 'yes';
	}

	# Handle is suspended
	else
	{
		$default   = $cgi->checkbox(-name=>'PUPDEF', -label=>'');
		$s_default = 'no';
	}


	# Display to realm
	# Realm display
	if (defined($param->{'ROOM'}))
	{
		# Display Puppeteer information
		my $wiget = (defined($param->{'REALM_DISPLAY'}))
		  ? $cgi->checkbox(-name=>'REALM_DISPLAY', -checked=>'on', -label=>'') 
		  : $cgi->checkbox(-name=>'REALM_DISPLAY', -label=>''); 


		# Final replace
		$preview  = $tmpl->Pass('TmplPupConfigPreview',
				WPUPPREV   => $wiget
		 );
	}

	
	# Title bar
	# Controls
	$title = ($database->{'USR'}{'Javascript'} ne 'yes')
	  ? $tmpl->Pass('TagPupTitleDef')
	  : $tmpl->Pass('TmplFormat', 
			MFUNC => 'strTitle',
			MNAME => $config->{'TagPupTitle'}
		 ); 

	# Widget
	$wtitle = $cgi->textarea(
		-name=>'PUPTIT',
		-rows=>3,
		-columns=>65,
		-wrap=>'SOFT',
		-default=>$info->{'PuppetTitle'});


	# Tagline bar
	# Controls
	$tag = ($database->{'USR'}{'Javascript'} ne 'yes')
	  ? $tmpl->Pass('TagPupTagDef')
	  : $tmpl->Pass('TmplFormat',
			MFUNC => 'strTagline',
			MNAME => $config->{'TagPupTagline'}
		 ); 

	# Physical widget
	$wtag = $cgi->textarea(
		-name=>'PUPTAG',
		-rows=>3,
		-columns=>65,
		-wrap=>'SOFT',
		-default=>$info->{'PuppetTag'}
	 );

	# Finish statement
	$statement->finish();


	# Security signature
	# Generate signature
	$sig = md5_hex(
		$info->{'PuppetPic'} .
	 	$info->{'PuppetBio'} .
		$gummy .
		$s_active .
		$s_default .
		$info->{'PuppetTitle'} .
		$info->{'PuppetTag'}
	 );


	# Information display
	# Start form
	print $cgi->start_form(-name=>'PUPPET');
	print $param->EmbedNormal($param->Flat(),
		PUPPET  => $param->{'PUPPET'},
		POOL    => $info->{'PuppetPool'},
	        MD5_SIG => $sig
	  ), "\n";


	# Display template
	$tmpl->Show('TmplPupConfig',
		LACTVT     => "HREF=\"$link->{URL}$link->{ACTIVITY}\"",
		LBACK      => "HREF=\"$link->{URL}$link->{BACK}\"",
		LCLONE     => "HREF=\"$link->{URL}$link->{CLONE}\"",
		LEXPRT     => "HREF=\"$link->{URL}$link->{EXPORT}\"",
		LIMPRT     => "HREF=\"$link->{URL}$link->{IMPORT}\"",
		LGROUP     => "HREF=\"$link->{URL}$link->{GROUP}\"",
		MNAME      => $info->{'PuppetName'},
		MPUPPREV   => $preview,
		MIMGWIDTH  => $info->{'PuppetPicWidth'},
		MIMGHEIGHT => $info->{'PuppetPicHeight'},
		MIMGSIZE   => $info->{'PuppetPicSize'},
		MPUPTITLEB => $title,
		MPUPTAGB   => $tag,
		WPUPIMG    => $cgi->textfield('PUPIMG', $info->{'PuppetPic'}, 30, 200),
		WPUPBIO    => $cgi->textfield('PUPBIO', $info->{'PuppetBio'}, 30, 200),
		WPUPGUMMY  => $gummies,
		WPUPACTV   => $active,
		WPUPDEF    => $default,
		WPUPTITLE  => $wtitle,
		WPUPTAG    => $wtag,
		WSUBMIT    => $cgi->submit($config->{'TxtSubmit'}),
		WRESET     => $cgi->reset($config->{'TxtReset'})
	 );

	# End of form
	print $cgi->end_form(), "\n";
}


#####################
# PuppetControlWrite
#
# Sub-routine that will handle the change of information for a puppet.  This
# is used as a sub process to handle database interaction.

sub PuppetControlWrite
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $param    = shift;					# Parameter handle
	my $config   = shift;					# Config hash

	my $audit;						# Auditing handler
	my $comm;						# Post handler
	my $filter   = new Ethereal::Filter();			# Filter handle
	my $tmpl;						# Template

	my $legal;						# Image legal
	my $height;						# Image height
	my $size;						# Image size
	my $width;						# Image width

	my $cbio;						# Bio
	my $cgum;						# Gummy
	my $ctag;						# Cleaned tag
	my $ctit;						# Cleaned title

	my $active;						# Active flag
	my $default;						# Default toggle
	my $file;						# Filename
	my $name;						# Name of gummy
	my $updates;						# Updates made in last hour

	my $lang;						# Language filter
	my @lang;						# Array of banned words

	my %realm;						# Realm hash

	#####################
	# Program area

	# Check for updates
	($updates) = $database->DataGetPuppetUpdates($param->{'USER'});

	# Updates are permitted
	if ($updates < $database->{'SYS'}{'SetThrottle'})
	{
		# Determine active state
		($active) = $database->DataGetPuppetActive($param->{'PUPPET'}, $param->{'USER'});


		# Image handling
		# If valid URL
		if ((length($param->{'PUPIMG'}) > 10)
		 && ($param->{'PUPIMG'} !~ /^http:\/\/\d{1,3}\.\d{1,3}\.\d{1,3}/i)) 
		{
			# If legal
			if (ImageLegal($database, $param->{'PUPIMG'}))
			{	
				# Capture and determin size
				($width, $height, $size, $_) = ImageSize($database, $param->{'PUPIMG'}); 

				# Make sure image is valid
				$param->{'PUPIMG'} = defined($width) ? $param->{'PUPIMG'} : undef;

	
				# Update database with new information
				$database->Quick("UPDATE Puppet
					SET PuppetPic=?, 
					    PuppetPicWidth=?, 
					    PuppetPicHeight=?,
					    PuppetPicSize=?
					WHERE PuppetName=? 
					  AND PuppeteerLogin=?", 
	
				 $param->{'PUPIMG'}, 
				 $width, 
				 $height, 
				 $size,
				 $param->{'PUPPET'}, 
				 $param->{'USER'}
				); 
			}
		}
		else
		{
			# Update database with nulls
			$database->Quick("UPDATE Puppet 
				SET PuppetPic=NULL, 
			            PuppetPicWidth=NULL, 
				    PuppetPicHeight=NULL,
				    PuppetPicSize=NULL
				WHERE PuppetName=? 
				  AND PuppeteerLogin=?", 

			 $param->{'PUPPET'}, 
			 $param->{'USER'}
			); 
		}

		# Assign default
		if (defined($param->{'PUPDEF'}))
		{
			# Database updates
			# Pull any stray tags
			$database->Quick("UPDATE Puppet
				SET PuppetDefault='no'
				WHERE PuppeteerLogin=?
				  AND PuppetDefault='yes'",
	
			 $param->{'USER'}
			);

			# Assign proper default
			$database->Quick("UPDATE Puppet
				SET PuppetDefault='yes'
				WHERE PuppetName=?
				  AND PuppeteerLogin=?
				  AND PuppetPool='system'",
	
			 $param->{'PUPPET'},
			 $param->{'USER'}
			);
		}


		# Active puppets
		# Activate
		if ((defined($param->{'PUPACT'})) 
		 && (defined($active)))
		{
			# Remove timestamp
			$database->Quick("UPDATE Puppet
				SET PuppetActive=NULL
				WHERE PuppetName=?
				  AND PuppeteerLogin=?",

			 $param->{'PUPPET'},
			 $param->{'USER'}
			);
		}

		# Deactivate
		elsif ((!defined($param->{'PUPACT'}))
		 && (!defined($active)))
		{
			# Remove timestamp
			$database->Quick("UPDATE Puppet
				SET PuppetActive=LOCALTIMESTAMP
				WHERE PuppetName=?
				  AND PuppeteerLogin=?",

			 $param->{'PUPPET'},
			 $param->{'USER'}
			);
		}


		# Data handling
		# Truncate and clean output
		$ctag = substr($param->{'PUPTAG'}, 0, 1024);
		$ctit = substr($param->{'PUPTIT'}, 0, 1024);
		$cbio = $param->{'PUPBIO'};


		# Gummy handling
		($cgum) = ($param->{'GUMMY'} ne $config->{'TxtNone'})
		  ? $database->DataGetGummyPath($param->{'GUMMY'})
		  : 'NIL';	

		# Tag handling
		($ctag) = $filter->SafeHTML($ctag, $config->{'SetAbuse'});
		($ctit) = $filter->SafeHTML($ctit, $config->{'SetAbuse'});

		# Language filter
		# Retreive all words
		$database->GetList(\@lang, "SELECT RestrictionValue
			FROM  Restriction
			WHERE RestrictionType='word'
			  AND RealmName='system'
			GROUP BY RestrictionValue");

		# Language filtering
		while ($lang = shift(@lang))
		{
			# Avoid host stupidity
			$lang = quotemeta($lang);

			# Run a global replace
			$ctag =~ s/$lang/****/gsi;
			$ctit =~ s/$lang/****/gsi;
		}


		# Tagline, Title and Bio link
 		$ctag = (length($ctag) > 3)  ? $ctag : undef;
 		$ctit = (length($ctit) > 3)  ? $ctit : undef;
 		$cbio = ((length($cbio) > 12) && ($cbio =~ /^http:\/\//is)) ? $cbio : undef; 


		# Create instance
		$audit    = new Ethereal::Audit($database, $cgi, $param);


		# Database update
		eval
		{
			# Update fields
			$database->Write("UPDATE Puppet 
				SET PuppetTag=?, 
				    PuppetTitle=?,
				    PuppetBio=?,
				    PuppetGummy=?
				WHERE PuppetName=? 
				  AND PuppeteerLogin=?",
	
			 $ctag, 
			 $ctit, 
			 $cbio,
			 $cgum, 
			 $param->{'PUPPET'}, 
			 $param->{'USER'}
			);

			# Auditing entry
			$audit->PuppetMod($param->{'PUPPET'});

			# Commit changes
			$database->Commit();
		};

		# If something failed
		if ($@)
		{
			# Issue warning
			warn("Puppet modification aborted: $@");
	
			# Undo incomplete changes
			$database->Rollback();
		}


		# Display
		if (defined($param->{'REALM_DISPLAY'}))
		{
			# Link with hash
			$database->GetHashRealm(\%realm);
			
			# Define instance
			$tmpl  = new Ethereal::Template(\%realm);
			$comm  = new Ethereal::Comm($database, $cgi, $tmpl);
	
			# Pull values
			my ($tsize,
			    $tcolor,
			    $tface) = $database->DataGetFont($param->{'PUPPET'}, $param->{'USER'});
	
			# Basic font setup
			$param->{'SIZE'}  = $tsize;
			$param->{'COLOR'} = $tcolor;
			$param->{'FACE'}  = $tface;
	
			# Identifying marks
			$param->{'CHAR'}    = $param->{'PUPPET'};
			$param->{'SOURCE'}  = $param->{'USER'}; 
			$param->{'TARGET'}  = $param->{'USER'}; 
			$param->{'OTARGET'} = $param->{'PUPPET'}; 
	
			# Set post
			$param->{'POST'}    = $config->{'MsgPupPrintout'};
	
			# Send post
			$comm->DataWrite($param);
		}

		# Cleanup
		$param->Cleanup(
		  'CHAR',
		  'COLOR',
		  'FACE',
		  'OTARGET',
		  'POOL',
		  'POST',
		  'PUPACT',
		  'PUPDEF',
		  'PUPDIS',
		  'SOURCE',
		  'TARGET'
		 );
	}
}


#####################
# Puppet Gummy
#
# This function is a variant of the GetList function used to pull
# fields from the database.  However, since this one will not
# clean up the array which is generally desired. 

sub PuppetGummy
{
	#####################
	# Data members
	my $database = shift;					# Database handle

	my $statement;						# Database statement

	my $array    = shift;					# Reference to array
        my $class    = shift;					# Identification class                             
	my $realm    = shift;					# Realm name

	my @realms;						# Additional realm


	#####################
	# Program Area
	
	# Initial assignment
	# System realm
	push(@realms, 'system');


	# Realm gummies
	# Define realm
	if (defined($realm))
	{
		# Add additional realm
		push(@realms, $realm);

		# Define additional search
		$realm = 'OR RealmName=?';
	}

	# Realm
	else
	{
		# Blank it out
		$realm = '';
	}


	# Prepare and execute
	$database->Pull(\$statement, "SELECT GummyName 
		FROM Gummy 
		WHERE GummyClass=?
		  AND RealmName=? $realm
		ORDER BY GummyIdent, 
			 GummyName",

	 $class,
	 @realms
	);


	# Retrieve and place in array
	while ($_ = $statement->fetchrow())
	{ push(@{$array}, $_); }

	# End statement
	$statement->finish;
}

# Alias for administrators
sub GetGummyAdmin
{ PuppetGummy(shift, shift, 'admin'); }

# Alias for default gummy
sub GetGummyDefault
{ PuppetGummy(shift, shift, 'default'); }

# Alias for canon handles
sub GetGummyCanon
{ PuppetGummy(shift, shift, 'canon', shift); }

# Alias for contributor gummies
sub GetGummyContrib
{ PuppetGummy(shift, shift, 'contributor'); }

# Alias for general
sub GetGummyGeneral
{ PuppetGummy(shift, shift, 'general'); }

# Alias for moderators
sub GetGummyMod
{ PuppetGummy(shift, shift, 'mod', shift); }

# Alias for supervisors
sub GetGummySuper
{ PuppetGummy(shift, shift, 'super', shift); }
